'''
Click command function 'dump'.
'''

import time
import zipfile
import zlib

import click

from posce.comms.base import group
from posce            import tools

@group.command(short_help='Archive notes.')
@click.argument('file')
@click.option('-l', '--level',
    help    = 'Compression level (0-9).',
    default = 5,
    metavar = 'INT',
    type    = click.IntRange(0, 9),
)
@click.pass_obj
def dump(book, file, level):
    '''
    Create notes archive at FILE.
    '''

    opts = {
        'compression':   zipfile.ZIP_DEFLATED,
        'compresslevel': level,
    }

    with zipfile.ZipFile(file, 'w', **opts) as zipf:
        for note in book:
            zipf.write(note.path, note.base)



'''
Click base group function.
'''

import click

from posce            import VERSION_STRING
from posce.items      import Book
from posce.tools.clui import CustomGroup

SETTINGS = {
    'help_option_names': ['-h', '--help'],
}

def version(ctx, param, value):
    '''
    Print the current version.
    '''

    if value:
        click.echo(VERSION_STRING)
        ctx.exit()

@click.group(cls=CustomGroup, context_settings=SETTINGS)
@click.option('--dir', envvar='POSCE_DIR', hidden=True)
@click.option('--ext', envvar='POSCE_EXT', hidden=True)
@click.option('-v', '--version',
    help         = 'Show current version and exit.',
    callback     = version,
    expose_value = False,
    is_eager     = True,
    is_flag      = True,
)
@click.pass_context
def group(ctx, dir, ext):
    '''
    Posce: a note-taking toolkit for your command line.
    See github.com/posce/posce for help and issues.
    '''

    ctx.obj = Book(dir, ext.lstrip('.'))



'''
File I/O and discovery functions.
'''

import glob
import os
import shutil

OPTIONS = {
    'encoding': 'utf-8',
}

def append(path, string, *, sep='\n'):
    '''
    Append a string to a file.
    '''

    with open(path, 'a', **OPTIONS) as file:
        file.write(sep + string)

def copy(path, name):
    '''
    Copy a file to another name in the same directory.
    '''

    dire = os.path.dirname(path)
    ext  = os.path.splitext(path)[-1]
    dest = os.path.join(dire, name + ext)
    shutil.copyfile(path, dest)

def create(path, string):
    '''
    Write a string to a new file.
    '''

    with open(path, 'x', **OPTIONS) as file:
        file.write(string)

def exists(path):
    '''
    Return true if a file or directory exists.
    '''

    return os.path.exists(path)

def find(dire, term):
    '''
    Yield all files in a directory with names matching a glob pattern.
    '''

    pattern = os.path.normpath(os.path.join(dire, term))
    yield from glob.iglob(pattern)

def read(path):
    '''
    Return the contents of a file as a string.
    '''

    with open(path, 'r', **OPTIONS) as file:
        return file.read()

def reext(path, ext):
    '''
    Move a file to a different extension in the same directory.
    '''

    dire = os.path.dirname(path)
    name = os.path.splitext(os.path.basename(path))[0]
    dest = os.path.join(dire, f'{name}.{ext}')
    shutil.move(path, dest)

def rename(path, name):
    '''
    Move a file to a different name in the same directory.
    '''

    dire = os.path.dirname(path)
    ext  = os.path.splitext(path)[-1].lstrip('.')
    dest = os.path.join(dire, f'{name}.{ext}')
    shutil.move(path, dest)

def write(path, string):
    '''
    Write a string to a file.
    '''

    with open(path, 'w', **OPTIONS) as file:
        file.write(string)



'''
Filepath string manipulation functions.
'''

import os.path

EXPANSIONS = {
    '%': os.path.expandvars,
    '~': os.path.expanduser,
    '$': os.path.expandvars,
}

def base(path):
    '''
    Return a path's basename with the extension.
    '''

    return os.path.basename(path)

def clean(path):
    '''
    Return a normalised path.
    '''

    return os.path.normpath(path)

def expand(path):
    '''
    Return a clean variable-expanded path.
    '''

    for char, func in EXPANSIONS.items():
        if char in path:
            path = func(path)
    return path

def ext(path):
    '''
    Return a path's extension without a dot.
    '''

    part = os.path.splitext(path)
    return part[-1].lstrip('.')

def join(*elems):
    '''
    Return a clean joined filepath.
    '''

    return os.path.normpath(os.path.join(*elems))

def name(path):
    '''
    Return a path's basename without the extension.
    '''

    base = os.path.basename(path)
    return os.path.splitext(base)[0]

def parent(path):
    '''
    Return a path's parent directory.
    '''

    return os.path.normpath(os.path.dirname(path))



'''
Click command function 'edit'.
'''

import click

from posce.comms.base import group
from posce            import tools

@group.command(short_help='Edit note.')
@click.argument('name')
@click.option('-e', '--editor',
    help    = 'Editor program to use.',
    metavar = 'PROG',
    type    = str,
)
@click.pass_obj
def edit(book, name, editor):
    '''
    Open note NAME in editor.
    '''

    note = tools.clui.disambiguate(book, name)
    func = lambda note: click.edit(
        editor    = editor or None,
        extension = f'.{note.ext}',
        filename  = note.path,
    )

    if data := func(note):
        note.write(data)



'''
Click command function 'drop'.
'''

import click
import send2trash

from posce.comms.base import group
from posce            import tools

@group.command(short_help='Trash note.')
@click.argument('name')
@click.pass_obj
def drop(book, name):
    '''
    Move note NAME to trash.
    '''

    note = tools.clui.disambiguate(book, name)
    send2trash.send2trash(note.path)



'''
Click command function 'make'.
'''

import click

from posce.comms.base import group
from posce            import tools

@group.command(short_help='Create note.')
@click.argument('name')
@click.option('-f', '--file',
    help    = 'Copy note from FILE.',
    default = '',
    metavar = 'FILE',
)
@click.pass_obj
def make(book, name, file):
    '''
    Create new note NAME.
    '''

    if name in book:
        tools.clui.error(f'Note {name!r} already exists.')
    else:
        if file:
            try:
                path = tools.path.clean(file)
                data = tools.file.read(path)
                book.create(name, data)
            except Exception:
                tools.clui.error(f'Cannot read file {file!r}.')
        else:
            book.create(name, '')



'''
Command-line user interface functions.
'''

import click

DYM_LIMIT = 5

class CustomGroup(click.Group):
    '''
    A custom Click Group that disambiguates Command names.
    '''

    def get_command(self, ctx, name):
        '''
        Return a Command matching an unambiguous name.
        '''

        comms = [
            comm for comm in self.list_commands(ctx)
            if comm.startswith(name)
        ]

        if len(comms) == 0:
            return None
        elif len(comms) == 1:
            return click.Group.get_command(self, ctx, comms[0])
        else:
            clist = ', '.join(map(repr, comms))
            ctx.fail(f'Ambiguous command. Did you mean: {clist}?')


def disambiguate(book, name):
    '''
    Return a Note matching an unambiguous name, or raise an error if
    there are zero or multiple matching Notes.
    '''

    notes = list(book.disambiguate(name))

    if len(notes) == 0:
        error(f'Note {name!r} does not exist.')

    elif len(notes) == 1:
        return notes[0]

    elif len(notes) <= DYM_LIMIT:
        names = sorted(note.name for note in notes)
        nlist = ', '.join(map(repr, names))
        error(f'Ambiguous note name. Did you mean: {nlist}?')

    else:
        error(f'Ambiguous note name, {len(notes)} notes match.')

def error(string):
    '''
    Raise a ClickException with an error string.
    '''

    raise click.ClickException(string)



'''
Click command function 'find'.
'''

import re

import click

from posce.comms.base import group

@group.command(short_help='Search notes.')
@click.argument('term')
@click.option('-r', '--regex',
    help    = 'Use term as regular expression.',
    is_flag = True,
)
@click.pass_obj
def find(book, term, regex):
    '''
    List all notes matching TERM.
    '''

    term = term if regex else re.escape(term)
    for note in book.search(term):
        click.echo(note.name)



'''
Click command function 'show'.
'''

import textwrap

import click

from posce.comms.base import group
from posce            import tools

@group.command(short_help='Print note.')
@click.argument('name')
@click.option('-w', '--wrap',
    help    = 'Wrap text to column width.',
    default = 0,
    metavar = 'COLS',
    type    = int,
)
@click.pass_obj
def show(book, name, wrap):
    '''
    Print contents of note NAME.
    '''

    note = tools.clui.disambiguate(book, name)
    if wrap:
        for line in textwrap.wrap(note.read(), width=wrap, tabsize=4):
            click.echo(line)
    else:
        click.echo(note.read())



'''
Package definition for 'posce.comms'.
'''

from posce.comms import base
from posce.comms import clip
from posce.comms import copy
from posce.comms import drop
from posce.comms import dump
from posce.comms import edit
from posce.comms import find
from posce.comms import list
from posce.comms import make
from posce.comms import move
from posce.comms import repl
from posce.comms import show
from posce.comms import wget



'''
Click command function 'clip'.
'''

import click
import pyperclip

from posce.comms.base import group
from posce            import tools

@group.command(short_help='Clipboard note.')
@click.argument('name')
@click.pass_obj
def clip(book, name):
    '''
    Copy note NAME to clipboard.
    '''

    note = tools.clui.disambiguate(book, name)
    pyperclip.copy(note.read())



'''
Click command function 'copy'.
'''

import click

from posce.comms.base import group
from posce            import tools

@group.command(short_help='Copy note.')
@click.argument('name')
@click.argument('dest')
@click.pass_obj
def copy(book, name, dest):
    '''
    Copy note NAME to new note DEST.
    '''

    note = tools.clui.disambiguate(book, name)

    if dest in book:
        tools.clui.error(f'Note {dest!r} already exists.')
    else:
        note.copy(dest)



'''
Click command function 'repl'.
'''

import code

import click

from posce            import VERSION_STRING
from posce.comms.base import group

@group.command(hidden=True, short_help='Open REPL.')
@click.pass_obj
def repl(book):
    '''
    Open a debug REPL.
    '''

    args = {'book': book, 'group': group}
    code.interact(banner=VERSION_STRING, local=args, exitmsg='')



'''
Click command function 'wget'.
'''

import urllib.parse
import urllib.request

import click

from posce            import tools
from posce.comms.base import group

@group.command(short_help='Download note.')
@click.argument('name')
@click.argument('url')
@click.pass_obj
def wget(book, name, url):
    '''
    Download URL into note NAME.
    '''

    note = tools.clui.disambiguate(book, name)
    url  = url if url.startswith('http') else f'http://{url}'

    try:
        resp = urllib.request.urlopen(url)
        if data := resp.read():
            note.write(data.decode('utf-8'))

    except urllib.error.URLError:
        tools.clui.error(f'Could not find {url!r}.')



'''
Click command function 'move'.
'''

import click

from posce            import tools
from posce.comms.base import group

@group.command(short_help='Move note.')
@click.argument('name')
@click.argument('dest')
@click.pass_obj
def move(book, name, dest):
    '''
    Move note NAME to new note DEST.
    '''

    note = tools.clui.disambiguate(book, name)

    if dest in book:
        tools.clui.error(f'Note {dest!r} already exists.')
    else:
        note.rename(dest)
